const INVITE_HEADERS = "MIME-Version: 1.0\r\n"
  + "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n\r\n";
const MSNFTP_GUID = "{5D3E02AB-6190-11d3-BBBB-00C04F795683}";

function Invitation(aPassport, aSbSession) {
  this.passport = aPassport;
  this.sbSession = aSbSession;
  this.fileTransfer = null;
}

Invitation.prototype.acceptReceived = function(aAuthCookie, aIp, aSenderConnect) {
  var authCookie = 0;
  if (this.fileTransfer.getSend() && !aSenderConnect) {
    // We need to send a message with the offer to serve, then listen
    authCookie = Math.round(Math.random() * 1000000);
    var myIP = this.sbSession.getIPAddress();
    var invCookie = this.fileTransfer.getInvCookie();
    this.sendAcceptMessage(myIP, invCookie, authCookie, false);
  }
  else {
    authCookie = aAuthCookie;
  }
  this.fileTransfer.setAuthCookie(authCookie);
  this.fileTransfer.open(aIp);
}

Invitation.prototype.cancelFileTransfer = function() {
  this.fileTransfer.cancel();
  var invCookie = this.fileTransfer.getInvCookie();
  this.sendCancelMessage(invCookie, "TIMEOUT");
}

Invitation.prototype.getPassport = function() {
  return this.passport;
}

Invitation.prototype.fileTransferDone = function(aResult, aFileName) {
  if (aResult == 2) {
    var invCookie = this.fileTransfer.getInvCookie();
    this.sendCancelMessage(invCookie, "FTTIMEOUT");
  }
  delete this.fileTransfer;
  this.sbSession.showFileTransferResult(aResult, aFileName);
}

Invitation.prototype.invite = function(aFile, aBehindRouter) {
  var invCookie = Math.round(Math.random() * 1000000);
  this.fileTransfer = new MSNFTP(invCookie, 0, aFile, true, this);
  var payload = INVITE_HEADERS;
  payload += "Application-Name: File Transfer\r\n";
  payload += "Application-GUID: " + MSNFTP_GUID + "\r\n";
  payload += "Invitation-Command: INVITE\r\n";
  payload += "Invitation-Cookie: " + invCookie + "\r\n";
  payload += "Application-File: " + aFile.leafName + "\r\n";
  payload += "Application-FileSize: " + aFile.fileSize + "\r\n";
  if (aBehindRouter) {
    payload += "Connectivity: N\r\n";
  }
  this.sbSession.sendNMessage(payload);
}

Invitation.prototype.inviteReceived = function(aInvCookie, aFileName, aFileSize, aConnectivity) {
  var file = this.sbSession.acceptFile(aFileName, aFileSize);
  if (file) {
    if (!aConnectivity) {
      var myIP = this.sbSession.getIpAddress();
      this.sendAcceptMessage(myIP, aInvCookie, 0, false);
      this.fileTransfer = new MSNFTP(aInvCookie, 0, file, false, this);
    }
    else {
      var authCookie = Math.round(Math.random() * 1000000);
      this.sendAcceptMessage(aInvCookie, authCookie, true);
      this.fileTransfer = new MSNFTP(aInvCookie, authCookie, file, false, this);
      this.fileTransfer.open(null);
    }
  }
  else {
    this.sendCancelMessage(aInvCookie, "REJECT");
  }
}

Invitation.prototype.messageReceived = function(aFrom, aMsg) {
  var headers = Utils.parseHeaders(aMsg);
  var invCookie = headers["invitation-cookie"];
  switch (headers["invitation-command"]) {
    case "INVITE":
      if (headers["application-guid"] == MSNFTP_GUID) {
        var fileName = headers["application-file"];
        var fileSize = headers["application-filesize"];
        var connectivity = headers["connectivity"];
        this.inviteReceived(invCookie, fileName, fileSize, connectivity);
      }
      else {
        this.sendCancelMessage(invCookie, "REJECT_NOT_INSTALLED");
        this.sbSession.invitationUnsupported();
      }
      break;
    case "ACCEPT":
      if (this.fileTransfer) {
        var currentInvCookie = this.fileTransfer.getInvCookie();
        if (currentInvCookie == invCookie) {
          var authCookie = headers["authcookie"];
          var ip = headers["ip-address"];
          var senderConnect = headers["sender-connect"];
          this.fileTransferAcceptReceived(authCookie, ip, senderConnect);
        }
      }
      break;
    case "CANCEL":
      if (this.fileTransfer) {
        var currentInvCookie = this.fileTransfer.getInvCookie();
        if (currentInvCookie == invCookie) {
          this.fileTransfer.cancel(false);
          delete this.fileTransfer;
        }
      }
      break;
  }
}

Invitation.prototype.sendAcceptMessage = function(aMyIp, aInvCookie, aAuthCookie, aSenderConnect) {
  var payload = INVITE_HEADERS;
  if (aAuthCookie) {
    payload += "IP-Address: " + aMyIp + "\r\n"
      + "Port: 6891\r\n"
      + "PortX: 11178\r\n"
      + "AuthCookie: " + aAuthCookie + "\r\n";
  }
  if (senderConnect) {
    payload += "Sender-Connect: TRUE\r\n";
  }
  payload += "Invitation-Command: ACCEPT\r\n"
    + "Invitation-Cookie: " + aInvCookie + "\r\n"
    + "Launch-Application: FALSE\r\n"
    + "Request-Data: IP-Address:\r\n";
  this.sbSession.sendNMessage(payload);
}

Invitation.prototype.sendCancelMessage = function(aInvCookie, aCancelCode) {
  var payload = INVITE_HEADERS
    + "Invitation-Command: CANCEL\r\n"
    + "Invitation-Cookie: " + aInvCookie + "\r\n"
    + "Cancel-Code: " + aCancelCode + "\r\n";
  this.sbSession.sendNMessage(payload);
}

Invitation.prototype.showFileTransferProgress = function(aPercentage) {
  this.sbSession.showFileTransferProgress(aPercentage);
}

Invitation.prototype.showFileTransferStatus = function(aLabel) {
  this.sbSession.showFileTransferStatus(aLabel);
}