const PORT = 6891;

function MSNFTP(aInvCookie, aAuthCookie, aFile, aSend, aParent) {
  this.invCookie = aInvCookie;
  this.authCookie = aAuthCookie;
  this.file = aFile;
  this.send = aSend;
  this.parent = aParent;
  this.result = -1;
  this.connection = new BaseConnection(this);
  this.parent.showFileTransferStatus("waiting");
  this.parent.showFileTransferProgress(0);
}

MSNFTP.prototype.onProtocolMessageReceived = function(aMsg, aPayload) {
  switch (aMsg[0]) {
    case "VER":
      if (this.send) {
        this.connection.sendMsg(["VER", "MSNFTP"]);
      }
      else {
        this.connection.sendMsg([this.parent.getPassport(), this.authCookie]);
      }
      break;
    case "USR":
      if (aMsg[2] == this.authCookie) {
        this.connection.sendMsg(["FIL", this.file.fileSize]);
      }
      else {
        this.result = 2;
        this.close();
      }
      break;
    case "FIL":
      this.fileSize = aMsg[1];
      this.bytesLeft = this.fileSize;
      this.outputStream = Components.classes["@mozilla.org/network/safe-file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
      this.outputStream.init(this.file, 0x04 | 0x08 | 0x20, 0600, 0);
      this.binaryMode = true;
      this.connection.sendMsg(["TFR"]);
      this.parent.showFileTransferStatus("receiving");
      break;
    case "TFR":
      this.parent.showFileTransferStatus("sending");
      this.sendFile();
      break;
    case "BYE":
      if (aMsg[1] == "16777989") {
        this.parent.fileTransferDone(0, this.file.leafName);
      }
      break;
    case "CCL":
      this.result = 1;
      this.close();
      break;
  }
}

MSNFTP.prototype.onBinaryDataReceived = function(aBinData, aBinDataLength) {
  this.outputStream.write(aBinData, aBinDataLength);
  this.bytesLeft -= aBinDataLength;
  this.parent.showFileTransferProgress(Math.floor(((this.fileSize - this.bytesLeft) / this.fileSize) * 100));
  if (this.bytesLeft == 0) {
    if (this.outputStream instanceof Components.interfaces.nsISafeOutputStream) {
      this.outputStream.finish();
    }
    else {
      this.outputStream.close();
    }
    delete this.fileSize;
    delete this.bytesLeft;
    delete this.outputStream;
    this.connection.sendMsg(["BYE 16777989"]);
    this.result = 0;
    this.close();
  }
  else {
    if (this.result > 0) {
      this.close();
    }
  }
}

MSNFTP.prototype.onConnectionHeard = function() {
  this.connection.sendMsg(["VER", "MSNFTP"]);
}

MSNFTP.prototype.cancel = function() {
  this.result = 1;
  if (!this.connection.isConnected) {
    this.close();
  }
}

MSNFTP.prototype.close = function() {
  this.connection.disconnect();
  this.connection.stopListening();
  this.parent.fileTransferDone(this.result, this.file.leafName);
}

MSNFTP.prototype.getInvCookie = function() {
  return this.invCookie;
}

MSNFTP.prototype.getSend = function() {
  return this.send;
}

MSNFTP.prototype.open = function(aIp) {
  if (!aIp) {
    this.connection.listen(PORT);
  }
  else {
    this.connection.connect(aIp, PORT);
  }
  this.parent.showFileTransferStatus("connecting");
}

MSNFTP.prototype.sendFile = function() {
  var bytesLeft = this.file.fileSize;
  var inputStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
  var binaryStream = Components.classes["@mozilla.org/binaryinputstream;1"].createInstance(Components.interfaces.nsIBinaryInputStream);
  inputStream.init(this.file, -1, -1, false);
  binaryStream.setInputStream(inputStream);
  while (this.result == -1 && bytesLeft > 0) {
    var bytesToRead = 0;
    if (bytesLeft >= 2045) {
      bytesToRead = 2045;
    }
    else {
      bytesToRead = bytesLeft;
    }
    var header = String.fromCharCode(0, bytesToRead % 256, Math.floor(bytesToRead / 256));
    var bytes = binaryStream.readBytes(bytesToRead);
    bytesLeft -= bytesToRead;
    this.connection.sendData(header + bytes);
    this.parent.showFileTransferProgress(Math.floor(((this.file.fileSize - bytesLeft) / this.file.fileSize) * 100));
  }
  if (this.result > 0) {
    this.close();
  }
}

MSNFTP.prototype.setAuthCookie = function(aAuthCookie) {
  this.authCookie = aAuthCookie;
}