const FORWARD_LIST_NUMBER = 1;
const ALLOW_LIST_NUMBER = 2;
const BLOCK_LIST_NUMBER = 4;
const REVERSE_LIST_NUMBER = 8;
const NEW_CONTACT_LIST_FLAGS = 3;

function MsnSession(aPassport, aPassword, aStatus, aLocaleId, aController,
                    aMsnProtocol) {
  this.contacts = {};
  this.controller = aController;
  this.localeId = aLocaleId;
  this.msnObject = new MsnObject(aPassport);
  this.msnProtocol = aMsnProtocol;
  this.nUnreadEmails = 0;
  this.passport = aPassport;
  this.password = aPassword;
  this.platform = "win 4.00 i386";
  this.sbSessions = new Array();
  this.sessionLength = 0;
  this.status = aStatus;
  this.msnProtocol.setMsnSession(this);
  this.fillInMsnObject();
}

MsnSession.prototype.onConnectionClosed = function(aStatus) {
  this.controller.onLogOut();
}

MsnSession.prototype.onContactAlreadySignedIn
 = function(aEmail, aDisplayName, aStatus, aCapabilityFlags, aMsnObjectString) {
  var contact = this.getContact(aEmail);
  if (contact.status == "FLN") {
    if (contact.displayName != aDisplayName) {
      contact.displayName = aDisplayName;
      this.msnProtocol.updateContactDisplayName(contact.email,
                                                contact.displayName);
    }
    contact.status = aStatus;
    contact.capabilityFlags = aCapabilityFlags;
    if (aMsnObjectString && contact.msnObject.string != aMsnObjectString) {
      contact.msnObject.fillInFromString(aMsnObjectString);
    }
    this.controller.onContactAlreadySignedIn(contact.email);
    var blocked = this.isContactBlocked(contact.email);
    this.controller.onContactChanged(contact.email, contact.displayName,
                                     contact.status, blocked);
  }
}

MsnSession.prototype.onContactAddedToList = function(aEmail, aListName) {
  if (aListName == "RL") {
    if (!(aEmail in this.contacts)) {
      this.contacts[aEmail] = new MsnContact(aEmail, aEmail, 0);
    }
    var contact = this.contacts[aEmail];
    if (!this.isContactInList(aEmail, REVERSE_LIST_NUMBER)) {
      contact.listFlags += REVERSE_LIST_NUMBER;
    }
    if ((contact.listFlags & 14) == REVERSE_LIST_NUMBER) {
      if (this.gtc == "A" || this.blp == "BL") {
        openContactAddPrompt(contact, this);
      }
      else {
        contact.listFlags += ALLOW_LIST_NUMBER;
        var listsWindow = winMediator.getMostRecentWindow("msnmsgr:lists");
        if (listsWindow) {
          listsWindow.onContactAddedToList("AL", contact.email,
                                           contact.displayName);
        }
      }
    }
  }
}

MsnSession.prototype.onContactChanged
 = function(aEmail, aDisplayName, aStatus, aCapabilityFlags, aMsnObjectString) {
  var contact = this.getContact(aEmail);
  if (contact.displayName != aDisplayName) {
    contact.displayName = aDisplayName;
    this.msnProtocol.updateContactDisplayName(contact.email,
                                              contact.displayName);
  }
  var contactSignedIn = false;
  if (contact.status == "FLN") {
    contactSignedIn = true;
  }
  contact.status = aStatus;
  contact.capabilityFlags = aCapabilityFlags;
  var msnObjectChanged = false;
  if (aMsnObjectString && contact.msnObject.string != aMsnObjectString) {
    contact.msnObject.fillInFromString(aMsnObjectString);
    msnObjectChanged = true;
  }
  if (contactSignedIn) {
    this.controller.onContactSignedIn(contact.email, contact.displayName);
  }
  var blocked = this.isContactBlocked(contact.email);
  this.controller.onContactChanged(contact.email, contact.displayName,
                                   contact.status, blocked);
  this.broadcastContactChange(aEmail, msnObjectChanged);
}

MsnSession.prototype.onContactListReceived = function() {
  this.controller.onContactListReceived();
  this.changeStatus(this.status);
}

MsnSession.prototype.onContactReceived
 = function(aEmail, aDisplayName, aListFlags) {
  this.contacts[aEmail] = new MsnContact(aEmail, aDisplayName, aListFlags);
  var contact = this.contacts[aEmail];
  if (this.isContactOnForwardList(contact.email)) {
    var blocked = this.isContactBlocked(contact.email)
    this.controller.onContactAdded(contact.email, contact.displayName, blocked);
  }
  if ((contact.listFlags & 14) == REVERSE_LIST_NUMBER) {
    if (this.gtc == "A" || this.blp == "BL") {
      openContactAddPrompt(contact, this);
    }
    else {
      contact.listFlags += ALLOW_LIST_NUMBER;
    }
  }
}

MsnSession.prototype.onContactRemovedFromList = function(aEmail, aListName) {
  if (aListName == "RL") {
    if (aEmail in this.contacts
        && !this.isContactInList(aEmail, REVERSE_LIST_NUMBER)) {
      this.contacts[aEmail].listFlags -= REVERSE_LIST_NUMBER;
    }
  }
}

MsnSession.prototype.onContactSignedOut = function(aEmail) {
  var contact = this.getContact(aEmail);
  if (contact.status != "FLN") {
    contact.status = "FLN";
    var blocked = this.isContactBlocked(contact.email);
    this.controller.onContactSignedOut(contact.email, contact.displayName,
                                       blocked);
    this.broadcastContactChange(contact.email, false);
  }
}

MsnSession.prototype.onDisplayNameChanged = function(aDisplayName) {
  this.displayName = aDisplayName;
  this.controller.onDisplayNameChanged(this.displayName);
}

MsnSession.prototype.onEmailLogInReceived = function(aRru, aAction, aId) {
  var rru = aRru;
  if (this.mailToEmail) {
    rru += "?mailto=1&to=" + this.mailToEmail;
    delete this.mailToEmail;
  }
  var credsToHash = this.mspAuth + this.sessionLength + this.password;
  var fields = {
    login : this.passport.substr(0, this.passport.indexOf("@")),
    username : this.passport,
    sid : this.sid,
    kv : this.kv,
    id : aId,
    sl : this.sessionLength,
    rru : rru,
    auth : this.mspAuth,
    creds : Utils.getMd5HashFromStringAsHex(credsToHash),
    svc : "mail",
    js : "yes"
  };
  this.controller.openWebMail(aAction, fields);
}

MsnSession.prototype.onEmailMessagesRead = function(nEmails) {
  this.nUnreadEmails -= nEmails;
  this.controller.onNewEmailMessageCountChanged(this.nUnreadEmails);
}

MsnSession.prototype.onEmailMessagesRecovered = function(nEmails) {
  this.nUnreadEmails += nEmails;
  this.controller.onNewEmailMessageCountChanged(this.nUnreadEmails);
}

MsnSession.prototype.onEmailMessagesRemoved = function(nEmails) {
  this.nUnreadEmails -= nEmails;
  this.controller.onNewEmailMessageCountChanged(this.nUnreadEmails);
}

MsnSession.prototype.onErrorReceived = function(aErrorCode) {
  this.controller.onErrorReceived(aErrorCode);
}

MsnSession.prototype.onInitialEmailNotification = function(anUnreadEmails) {
  this.nUnreadEmails = anUnreadEmails;
  this.controller.onNewEmailMessageCountChanged(this.nUnreadEmails);
}

MsnSession.prototype.onLogInFailed = function(aErrorCode) {
  this.controller.onLogInFailed(aErrorCode);
}

MsnSession.prototype.onLogInProgress = function(aStep) {
  this.controller.onLogInProgress(aStep);
}

MsnSession.prototype.onLogInSuccessful = function() {
  var me = this;
  this.sessionLengthUpdater = setInterval(function() { me.sessionLength++; },
                                          1000);
  this.controller.onLogInSuccessful();
}

MsnSession.prototype.onNewEmailMessageReceived
 = function(aSender, aArrivedInInbox) {
  if (aArrivedInInbox) {
    this.nUnreadEmails++;
    this.controller.onNewEmailMessageCountChanged(this.nUnreadEmails);
  }
  this.controller.onNewEmailMessageReceived(aSender);
}

MsnSession.prototype.onStatusChanged = function(aStatusCode) {
  this.status = aStatusCode;
  this.controller.onStatusChanged(this.status);
}

MsnSession.prototype.addContactToForwardList = function(aEmail) {
  if (aEmail in this.contacts) {
    var contact = this.contacts[aEmail];
    contact.lists += FORWARD_LIST_NUMBER;
    var blocked = this.isContactBlocked(contact.email);
    this.controller.onContactAdded(contact.email, contact.displayName, blocked);
    this.msnProtocol.addContactToList(contact.email, contact.displayName, "FL");
  }
  else {
    this.contacts[aEmail] = new MsnContact(aEmail, aEmail,
                                           NEW_CONTACT_LIST_FLAGS);
    var contact = this.contacts[aEmail];
    this.controller.onContactAdded(contact.email, contact.displayName, false);
    this.msnProtocol.addContactToList(contact.email, contact.displayName, "AL");
    this.msnProtocol.addContactToList(contact.email, contact.displayName, "FL");
    var listsWindow = winMediator.getMostRecentWindow("msnmsgr:lists");
    if (listsWindow) {
      listsWindow.onContactAddedToList("AL", contact.email, contact.displayName);
    }
  }
}

MsnSession.prototype.addContactToPrivacyList = function(aEmail, aAllowContact) {
  var contact = this.contacts[aEmail];
  var list = aAllowContact ? "AL" : "BL";
  this.msnProtocol.addContactToList(contact.email, contact.displayName, list);
}

MsnSession.prototype.blockContact = function(aEmail) {
  var contact = this.contacts[aEmail];
  var blocked = this.isContactInList(contact.email, BLOCK_LIST_NUMBER);
  var removeList = "";
  var addList = "";
  var listsWindow = winMediator.getMostRecentWindow("msnmsgr:lists");
  if (!blocked) {
    removeList = "AL";
    addList = "BL";
  }
  else {
    removeList = "BL";
    addList = "AL";
  }
  contact.listFlags -= getListNumber(removeList);
  contact.listFlags += getListNumber(addList);
  this.msnProtocol.removeContactFromList(contact.email, removeList);
  this.msnProtocol.addContactToList(contact.email, contact.displayName,
                                    addList);
  this.controller.onContactChanged(contact.email, contact.displayName,
                                   contact.status, !blocked);
  if (listsWindow) {
    listsWindow.onContactRemovedFromList(removeList, contact.email);
    listsWindow.onContactAddedToList(addList, contact.email,
                                     contact.displayName);
  }
}

MsnSession.prototype.broadcastContactChange = function(aEmail, aMsnObjChanged) {
  for (var i in this.sbSessions) {
    this.sbSessions[i].onContactChanged(aEmail, aMsnObjChanged);
  }
}

MsnSession.prototype.changeDisplayName = function(aDisplayName) {
  this.msnProtocol.changeDisplayName(aDisplayName);
}

MsnSession.prototype.changeDisplayPicture = function() {
  this.fillInMsnObject();
  this.changeStatus(this.status);
}

MsnSession.prototype.changeStatus = function(aStatus) {
  this.msnProtocol.changeStatus(aStatus, this.msnObject.string);
}

MsnSession.prototype.deleteSbSession = function(sbSession) {
  for (var i in this.sbSessions) {
    if (this.sbSessions[i] == sbSession) {
      this.sbSessions.splice(i, 1);
      dump("sbSession deleted.\n");
    }
  }
}

MsnSession.prototype.fillInMsnObject = function() {
  var file = Utils.getExtensionSubDirectory("pictures");
  file.append(btoa(this.passport));
  if (file.exists()) {
    this.msnObject.fillInFromDisplayPicture(file);
  }
  else {
    delete this.msnObject;
    this.msnObject = new MsnObject(this.passport);
  }
}

MsnSession.prototype.getContact = function(aEmail) {
  return this.contacts[aEmail];
}

MsnSession.prototype.getListContacts = function(aListName) {
  var listContacts = {};
  var listNumber = getListNumber(aListName);
  for (var email in this.contacts) {
    var contact = this.contacts[email];
    if (this.isContactInList(contact.email, listNumber)) {
      listContacts[email] = contact;
    }
  }
  return listContacts;
}

MsnSession.prototype.getOnlineContacts = function() {
  var onlineContacts = {};
  for (var email in this.contacts) {
    var contact = this.contacts[email];
    if (contact.status != "FLN" && !this.isContactBlocked(contact.email)) {
      onlineContacts[email] = contact;
    }
  }
  return onlineContacts;
}

MsnSession.prototype.getSbSession = function(aEmail) {
  var sbSession = null;
  for (i in this.sbSessions) {
    if (this.sbSessions[i].isSessionForContact(aEmail)) {
      sbSession = this.sbSessions[i];
    }
  }
  return sbSession;
}

MsnSession.prototype.isContactDeletable = function(aEmail) {
  var contact = this.contacts[aEmail];
  var deletable = false;
  if (contact.listFlags == ALLOW_LIST_NUMBER
      || contact.listFlags == BLOCK_LIST_NUMBER) {
    deletable = true;
  }
  return deletable;
}

MsnSession.prototype.isContactInList = function(aEmail, aListNumber) {
  var isInList = false;
  if ((this.contacts[aEmail].listFlags & aListNumber) == aListNumber) {
    isInList = true;
  }
  return isInList;
}

MsnSession.prototype.isContactBlocked = function(aEmail) {
  return this.isContactInList(aEmail, BLOCK_LIST_NUMBER);
}

MsnSession.prototype.isContactOnForwardList = function(aEmail) {
  return this.isContactInList(aEmail, FORWARD_LIST_NUMBER);
}

MsnSession.prototype.logIn = function() {
  this.msnProtocol.logIn();
}

MsnSession.prototype.logOut = function() {
  this.msnProtocol.logOut();
  this.controller.onLogOut();
}

MsnSession.prototype.openEmailInbox = function() {
  this.msnProtocol.requestEmailInboxLogIn();
}

MsnSession.prototype.openEmailMessageComposeForm = function(aEmail) {
  this.mailToEmail = aEmail;
  this.msnProtocol.requestEmailMessageComposeLogIn(aEmail);
}

MsnSession.prototype.openSbSession = function(aEmail, aSb) {
  var sbSession = this.getSbSession(aEmail);
  if (sbSession) {
    if (aSb) {
      sbSession.answer(aSb.ip, aSb.port, aSb.cki, aSb.sessionId);
    }
    else {
      sbSession.show();
    }
  }
  else {
    var sbSession = new SbSession(aEmail, this);
    if (aSb) {
      sbSession.answer(aSb.ip, aSb.port, aSb.cki, aSb.sessionId);
    }
    else {
      sbSession.openWindow();
      this.msnProtocol.requestSbSession(sbSession);
    }
    this.sbSessions.push(sbSession);
  }
}

MsnSession.prototype.removeContactFromForwardList = function(aEmail) {
  this.contacts[aEmail].listFlags -= FORWARD_LIST_NUMBER;
  this.msnProtocol.removeContactFromList(aEmail, "FL");
  this.controller.onContactRemoved(aEmail);
}

MsnSession.prototype.removeDeletableContact = function(aEmail) {
  var contact = this.contacts[aEmail];
  var listName = (contact.listFlags == ALLOW_LIST_NUMBER) ? "AL" : "BL";
  delete this.contacts[aEmail];
  this.msnProtocol.removeContactFromList(aEmail, listName);
}

MsnSession.prototype.requestSbSession = function(aSbSession) {
  this.msnProtocol.requestSbSession(aSbSession);
}

function getListNumber(aListName) {
  var listNumber = 0;
  switch (aListName) {
    case "FL":
      listNumber = FORWARD_LIST_NUMBER;
      break;
    case "AL":
      listNumber = ALLOW_LIST_NUMBER;
      break;
    case "BL":
      listNumber = BLOCK_LIST_NUMBER;
      break;
    case "RL":
      listNumber = REVERSE_LIST_NUMBER;
      break;
  }
  return listNumber;
}

function openContactAddPrompt(aContact, aSession) {
  var url = "chrome://msnmsgr/content/contactaddprompt.xul";
  var features = "chrome,resizable";
  var argument = {
    email : aContact.email,
    displayName : aContact.displayName,
    session : aSession
  }
  window.openDialog(url, "_blank", features, argument);
}