function SbSession(aEmail, aMsnSession) {
  this.bundle = Utils.getStringBundle("chrome://msnmsgr/locale/session.properties");
  this.connection = new BaseConnection(this);
  this.msnSession = aMsnSession;

  // user
  this.contacts = new Array();
  this.contacts.push(this.msnSession.getContact(aEmail));

  // session
  this.incomingMsgBuffer = new Array();
  this.outgoingMsgBuffer = new Array();
  this.lastMsgRcvTime = null;
  this.sessionId = 0;
  this.sessionStatus = 1;
  this.trID = 0;

  // helpers
  this.dontSend = false;
  this.requestDisplayPicture = false;
  this.sendTimeout = null;
  this.sessionWindowStatus = 0;

  // handlers
  this.invitation = new Invitation(this.msnSession.passport, this);
  this.imLogger = null;
  this.p2p = new P2P(this.msnSession.passport, this);
}

SbSession.prototype.onMessageStyleChanged = function() {
  if (this.sessionWindowStatus == 2) {
    this.sessionWindow.onMessageStyleChanged(MessageStyle.myStyle);
  }
}

SbSession.prototype.onProtocolMessageReceived = function(aMsg, aPayload) {
  switch (aMsg[0]) {
    case "USR":
      if (aMsg[2] == "OK") {
        this.sendCalCommand(this.contacts[0].email);
      }
      break;
    case "CAL":
      if (aMsg[2] == "RINGING") {
        this.sessionId = aMsg[3];
      }
      break;
    case "JOI":
      var email = aMsg[1];
      var contact = this.msnSession.getContact(email);
      if (!contact) {
        var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[2]));
        contact = new MsnContact(email, displayName, 0);
      }
      if (contact.email != this.contacts[0].email) {
        this.contacts.push(contact);
      }
      this.sessionStatus = 2;
      this.updateContactInfo();
      if (this.contacts.length > 1) {
        this.showNotificationMessage("joinMessage", contact.displayName);
      }
      else if (this.requestDisplayPicture) {
        this.p2p.requestDisplayPicture(contact);
      }
      if (this.outgoingMsgBuffer.length > 0) {
        for (var i in this.outgoingMsgBuffer) {
          var msgObj = this.outgoingMsgBuffer[i];
          this.connection.sendMsg(msgObj.msg, msgObj.payload);
        }
        delete this.outgoingMsgBuffer;
        this.outgoingMsgBuffer = new Array();
      }
      break;
    case "IRO":
      var contact = this.msnSession.getContact(aMsg[4]);
      if (!contact) {
        var displayName = decodeURIComponent(Utils.convertUTF8BytesToUTF16(aMsg[5]));
        contact = new MsnContact(aMsg[4], displayName, 0);
      }
      if (contact.email != this.contacts[0].email) {
        this.contacts.push(contact);
      }
      else if (this.requestDisplayPicture) {
        this.p2p.requestDisplayPicture(contact);
      }
      break;
    case "ANS":
      this.sessionStatus = 2;
      this.updateContactInfo();
      break;
    case "BYE":
      var contact = this.getContact(aMsg[1]);
      if (this.contacts.length == 1) {
        this.sessionStatus = 0;
      }
      else {
        this.contacts.splice(this.getContactIndex(contact.email), 1);
        this.updateContactInfo();
        this.showNotificationMessage("leaveMessage", contact.displayName);
      }
      break;
    case "MSG":
      var msgHeadersEnd = aPayload.indexOf("\r\n\r\n");
      var msgHeaders = Utils.parseHeaders(aPayload.substr(0, msgHeadersEnd));
      var msgBodyLength = aPayload.length - msgHeadersEnd - 1;
      var msgBody = aPayload.substr(msgHeadersEnd + 4, msgBodyLength);
      this.onMessageReceived(aMsg[1], msgHeaders, msgBody);
      break;
    case "ACK":
      // message acknowledged
      break;
    case "NAK":
      this.showNotificationMessage("messageNotSent", "");
      break;
    case "217":
      this.showNotificationMessage("contactOffline", "");
      break;
    default:
      if (this.sessionWindowStatus == 2) {
        var msg = "Unexpected command: " + aMsg[0];
        this.sessionWindow.showNotificationMessage(msg)
      }
      break;
  }
}

SbSession.prototype.onConnectionClosed = function(aStatus) {
  if (this.contacts.length == 1) {
    this.sessionStatus = 0;
  }
  else {
    this.close();
    this.showNotificationMessage("closedMessage", "");
    this.sessionWindow.onSessionClosed();
  }
}

SbSession.prototype.onContactChanged = function(aEmail, aDisplayName, aStatus, aMsnObjectChanged) {
  var contact = this.getContact(aEmail);
  if (contact) {
    this.updateContactInfo();
  }
}

SbSession.prototype.onMessageReceived = function(aContactId, aHeaders, aBody) {
  switch (aHeaders["content-type"]) {
    case "text/plain; charset=UTF-8":
      var timeStamp = Utils.getTimeStamp("/", ":");
      var contact = this.getContact(aContactId);
      contact.timeout = 0;
      var mms = "";
      if ("x-mms-im-format" in aHeaders) {
        mms = aHeaders["x-mms-im-format"];
      }
      var style = "";
      if (mms) {
        style = MessageStyle.getStyleFromMms(mms);
      }
      this.lastMsgRcvTime = timeStamp.substr(0, timeStamp.length - 3);
      var msgText = Utils.convertUTF8BytesToUTF16(aBody);
      if (this.sessionWindowStatus == 2) {
        this.sessionWindow.showIM(contact.email, contact.displayName, msgText, style, false);
        this.updateStatusMessage();
        this.sessionWindow.getAttention();
      }
      else {
        var msg = { email: contact.email, displayName: contact.displayName, msgText: msgText, style: style };
        this.incomingMsgBuffer.push(msg);
        if (this.sessionWindowStatus == 0) {
          this.openWindow();
        }
      }
      if (this.imLogger) {
        this.imLogger.log(timeStamp, aContactId, aBody);
      }
      break;
    case "text/x-msmsgscontrol":
      if ("typinguser" in aHeaders) {
        var contact = this.getContact(aHeaders["typinguser"].toLowerCase());
        contact.timeout = new Date().getTime();
        var me = this;
        setTimeout(function() { me.updateStatusMessage(); }, 6000);
        this.updateStatusMessage();
      }
      break;
    case "text/x-msmsgsinvite; charset=UTF-8":
      this.invitation.messageReceived(aContactId, aBody);
      break;
    case "application/x-msnmsgrp2p":
      this.p2p.onMessageReceived(aContactId, aBody);
      break;
  }
}

SbSession.prototype.onSessionWindowLoaded = function() {
  this.sessionWindowStatus = 2;
  this.sessionWindow.setSession(this);
  this.sessionWindow.onMessageStyleChanged(MessageStyle.myStyle);
  for (i in this.incomingMsgBuffer) {
    var msg = this.incomingMsgBuffer[i];
    this.sessionWindow.showIM(msg.email, msg.displayName, msg.msgText, msg.style);
  }
  delete this.incomingMsgBuffer;
  this.updateContactInfo();
  this.updateStatusMessage();
  this.updateDisplayPicture();
}

SbSession.prototype.onSessionWindowUnloaded = function() {
  this.close();
  this.msnSession.deleteSbSession(this);
}

SbSession.prototype.onTyping = function() {
  if (this.sessionStatus == 0 && this.contacts.length == 1) {
    this.sessionStatus = 1;
    this.restoreSession();
  }
  if (this.sessionStatus == 2 && !this.dontSend) {
    this.dontSend = true;
    this.sendTypingMessage();
    var me = this;
    this.sendTimeout = setTimeout(function() { me.dontSend = false; }, 5000);
  }
}

SbSession.prototype.open = function(aHost, aPort, aCki) {
  this.connection.connect(aHost, aPort);
  this.connection.sendMsg(["USR", ++this.trID, this.msnSession.passport, aCki]);
}

SbSession.prototype.close = function() {
  if (this.imLogger) {
    this.turnLoggingOff();
  }
  if (this.sendTimeout) {
    clearTimeout(this.sendTimeout);
  }
  if (this.connection.isConnected) {
    this.connection.sendMsg(["OUT"]);
    this.connection.disconnect();
  }
}

SbSession.prototype.acceptFile = function(aFileName, aFileSize) {
  var file = null;
  var fileSizeInKB = Math.ceil(aFileSize / 1024);
  var fileSizeString = fileSizeInKB  + " " + this.bundle.GetStringFromName("kiloByte");
  var topic = this.bundle.GetStringFromName("fileTransfer");
  var question = this.bundle.GetStringFromName("acceptFile1").replace("%S", aFileName)
    + this.bundle.GetStringFromName("acceptFile2").replace("%S", fileSizeString);
  if (Utils.askUser(topic, question)) {
    var filePicker = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
    filePicker.init(window, "Choose where you want to save the file", nsIFilePicker.modeGetFolder);
    var filePickerResult = filePicker.show();
    if (filePickerResult == nsIFilePicker.returnOK) {
      file = filePicker.file;
      file.append(aFileName);
      this.sessionWindow.showFileTransferBox(true);
      this.showNotificationMessage("fileTransferInit", file.leafName);
    }
  }
  return file;
}

SbSession.prototype.answer = function(aHost, aPort, aCki, aSessionId) {
  this.sessionStatus = 1;
  if (this.connection.isConnected) {
    this.connection.sendMsg(["OUT"]);
    this.connection.disconnect();
  }
  this.reset();
  this.connection.connect(aHost, aPort);
  this.connection.sendMsg(["ANS", ++this.trID, this.msnSession.passport, aCki, aSessionId]);
  this.aSessionId = aSessionId;
}

SbSession.prototype.cancelFileTransfer = function() {
  this.invitation.cancelFileTransfer();
}

SbSession.prototype.getContact = function(aEmail) {
  var contact = null;
  for (i in this.contacts) {
    if (this.contacts[i].email == aEmail) {
      contact = this.contacts[i];
    }
  }
  return contact;
}

SbSession.prototype.getContactIndex = function(aEmail) {
  var index = -1;
  for (i in this.contacts) {
    if (this.contacts[i].email == aEmail) {
      index = i;
    }
  }
  return index;
}

SbSession.prototype.getIpAddress = function() {
  return this.msnSession.ipAddress;
}

SbSession.prototype.inviteContact = function() {
  var url = "chrome://msnmsgr/content/invitecontact.xul";
  var argument = {
    profile : this.msnSession,
    sbSession : this
  }
  window.openDialog(url, "_blank", "chrome", argument);
}

SbSession.prototype.invitationUnsupported = function() {
  this.showNotificationMessage("inviteUnsupported", "");
}

SbSession.prototype.isSessionForContact = function(aEmail) {
  var answer = false;
  if (this.contacts[0].email == aEmail && this.contacts.length == 1) {
    answer = true;
  }
  return answer;
}

SbSession.prototype.openLink = function(aHref) {
  var recentWindow = winMediator.getMostRecentWindow("navigator:browser");
  if (recentWindow) {
    var theBrowser = recentWindow.getBrowser();
    var addedTab = theBrowser.addTab(aHref);
    theBrowser.selectedTab = addedTab;
  }
  else {
    window.open(href);
  }
}

SbSession.prototype.openWindow = function() {
  this.sessionWindowStatus = 1;
  this.sessionWindow = window.open("chrome://msnmsgr/content/session.xul", "_blank", "chrome,resizable");
  var me = this;
  this.sessionWindow.addEventListener("load", function() { me.onSessionWindowLoaded(); }, false);
}

SbSession.prototype.reset = function() {
  this.sessionID = 0;
  this.trID = 0;
  this.dontSend = false;
  if (this.sendTimeout) {
    clearTimeout(this.sendTimeout);
  }
}

SbSession.prototype.restoreSession = function() {
  if (this.connection.isConnected) {
    this.sendCalCommand(this.contacts[0].email);
  }
  else {
    this.msnSession.requestSbSession(this);
  }
}

SbSession.prototype.sendCalCommand = function(aEmail) {
  this.connection.sendMsg(["CAL", ++this.trID, aEmail]);
}

SbSession.prototype.sendDMessage = function(aContactId, aMsgBody) {
  var payload = "MIME-Version: 1.0\r\n"
    + "Content-Type: application/x-msnmsgrp2p\r\n"
    + "P2P-Dest: " + aContactId + "\r\n"
    + "\r\n"
    + aMsgBody;
  var msg = ["MSG", ++this.trID, "D"];
  if (this.sessionStatus == 2) {
    this.connection.sendMsg(msg, payload);
  }
  else {
    var msgObj = { msg: msg, payload: payload };
    this.outgoingMsgBuffer.push(msgObj);
    if (this.sessionStatus == 0 && this.contacts.length == 1) {
      this.sessionStatus = 1;
      this.restoreSession();
    }
  }
}

SbSession.prototype.sendFile = function() {
  var filePicker = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
  filePicker.init(this.sessionWindow, this.bundle.GetStringFromName("chooseFile"), nsIFilePicker.modeOpen);
  var filePickerResult = filePicker.show();
  if (filePickerResult == nsIFilePicker.returnOK) {
    var file = filePicker.file;
    var msncVersion = Math.floor(this.contacts[0].capabilityFlags / 0x10000000);
    if (msncVersion > 0) {
      this.p2p.inviteForFileTransfer(this.contacts[0].email, file);
    }
    else {
      this.invitation.invite(file, behindRouter);
    }
    this.sessionWindow.showFileTransferBox(true);
    this.showNotificationMessage("fileTransferInit", file.leafName);
  }
}

SbSession.prototype.sendIM = function(aMsgText) {
  var mms = MessageStyle.myMms;
  var utf8MsgText = Utils.convertUTF16ToUTF8Bytes(aMsgText);
  var payload = "MIME-Version: 1.0\r\n"
    + "Content-Type: text/plain; charset=UTF-8\r\n"
    + "X-MMS-IM-Format: " + mms + "\r\n"
    + "\r\n"
    + utf8MsgText;
  this.sendNMessage(payload);
  var style = MessageStyle.myStyle;
  this.sessionWindow.showIM(this.msnSession.passport, this.msnSession.displayName, aMsgText, style, true);
}

SbSession.prototype.sendNMessage = function(aPayload) {
  var msg = ["MSG", ++this.trID, "N"];
  if (this.sessionStatus == 2) {
    this.connection.sendMsg(msg, aPayload);
  }
  else {
    var msgObj = { msg: msg, payload: aPayload };
    this.outgoingMsgBuffer.push(msgObj);
    if (this.sessionStatus == 0 && this.contacts.length == 1) {
      this.sessionStatus = 1;
      this.restoreSession();
    }
  }
}

SbSession.prototype.sendTypingMessage = function() {
  var payload = "MIME-Version: 1.0\r\n"
    + "Content-Type: text/x-msmsgscontrol\r\n"
    + "TypingUser: " + this.msnSession.passport + "\r\n"
    + "\r\n";
  this.connection.sendMsg(["MSG", ++this.trID, "U"], payload);
}

SbSession.prototype.setDisplayPicture = function(aFilePath) {
  if (this.sessionWindowStatus == 2) {
    this.sessionWindow.setDisplayPicture(aFilePath);
  }
}

SbSession.prototype.show = function() {
  if (this.sessionWindowStatus == 2) {
    this.sessionWindow.focus();
  }
  else if (this.sessionWindowStatus == 0) {
    this.openWindow();
  }
}

SbSession.prototype.showFileTransferProgress = function(aPercentage) {
  this.sessionWindow.setFileTransferProgress(aPercentage);
}

SbSession.prototype.showFileTransferResult = function(aResult, aFileName) {
  var resultLabel = "";
  switch (aResult) {
    case 0:
      resultLabel = "fileTransferComplete";
      break;
    case 1:
      resultLabel = "fileTransferCancelled";
      break;
    case 2:
      resultLabel = "fileTransferProblem";
      break;
  }
  this.sessionWindow.showFileTransferBox(false);
  this.showNotificationMessage(resultLabel, aFileName);
}

SbSession.prototype.showFileTransferStatus = function(aLabel) {
  var statusText = this.bundle.GetStringFromName(aLabel);
  this.sessionWindow.setFileTransferStatus(statusText);
}

SbSession.prototype.showNotificationMessage = function(aLabel, aReplaceStr) {
  if (this.sessionWindowStatus == 2) {
    var msg = this.bundle.GetStringFromName(aLabel);
    if (aReplaceStr) {
      msg = msg.replace("%S", aReplaceStr);
    }
    this.sessionWindow.showNotificationMessage(msg);
  }
}

SbSession.prototype.turnLoggingOff = function() {
  this.imLogger.closeLogFile();
  delete this.imLogger;
}

SbSession.prototype.turnLoggingOn = function() {
  var timeStamp = Utils.getTimeStamp("", "");
  this.imLogger = new IMLogger(timeStamp, this.contacts[0].email);
  var success = this.imLogger.openLogFile();
  if (success) {
    this.showNotificationMessage("logOn", "");
  }
  else {
    delete this.imLogger;
    this.showNotificationMessage("logFailed", "");
  }
}

SbSession.prototype.updateContactInfo = function() {
  if (this.sessionWindowStatus < 2) {
    return;
  }
  var contactsString = "";
  var statusColor = "";
  for (i in this.contacts) {
    var statusText = Utils.getStatusText(this.contacts[i].status);
    contactsString += this.contacts[i].displayName + " (" + statusText + ") <" + this.contacts[i].email + ">, ";
  }
  contactsString = contactsString.substr(0, contactsString.length - 2);
  if (this.contacts.length == 1 && this.contacts[0].status != "NLN") {
    statusColor = "gold";
  }
  else {
    statusColor = "lightblue";
  }
  this.sessionWindow.setTitleText(this.contacts[0].displayName);
  var talkingToString = this.bundle.GetStringFromName("sendMessageToLabel");
  var headerText = talkingToString + "  " + contactsString;
  this.sessionWindow.setHeaderText(headerText);
  this.sessionWindow.setHeaderColor(statusColor);
}

SbSession.prototype.updateDisplayPicture = function() {
  var file = Utils.getExtensionSubDirectory("pictures");
  var contact = this.contacts[0];
  file.append(btoa(contact.email));
  var mustRequest = false;
  if (contact.msnObject.type == 3) {
    if (file.exists()) {
      dump("Display picture file exists. Checking hash.\n");
      var currentFileHash = Utils.getSha1HashFromFileAsBinary(file);
      if (contact.msnObject.sha1d == currentFileHash) {
        dump("Hash matches. Setting display picture.\n");
        this.sessionWindow.setDisplayPicture("file:///" + file.target);
      }
      else {
        dump("Hash doesn't match.\n");
        mustRequest = true;
      }
    }
    else {
      dump("Display picture file doesn't exist.\n");
      mustRequest = true;
    }
  }
  if (mustRequest) {
    if (this.sessionStatus == 2) {
      dump("Requesting display picture.\n");
      this.p2p.requestDisplayPicture(contact);
    }
    else {
      this.requestDisplayPicture = true;
    }
  }
}

SbSession.prototype.updateStatusMessage = function() {
  if (this.sessionWindowStatus < 2) {
    return;
  }
  var typingContacts = "";
  var contactCount = 0;
  var time = new Date().getTime();
  for (i in this.contacts) {
    var contact = this.contacts[i];
    if (time <= contact.timeout + 5000) {
      contactCount++;
      typingContacts += contact.displayName + ", ";
    }
  }
  if (contactCount != 0) {
    typingContacts = typingContacts.substr(0, typingContacts.length - 2);
  }
  var statusMsg = "";
  switch (contactCount) {
    case 0:
      if (this.lastMsgRcvTime) {
        statusMsg = this.bundle.GetStringFromName("lastMessageReceived").replace("%S", this.lastMsgRcvTime);
      }
      else {
        statusMsg = this.bundle.GetStringFromName("noMessagesReceived");
      }
      break;
    case 1:
      var typingMsg = this.bundle.GetStringFromName("isTyping");
      statusMsg = typingMsg.replace("%S", typingContacts);
      break;
    default:
      var typingMsg = this.bundle.GetStringFromName("areTyping");
      statusMsg = typingMsg.replace("%S", typingContacts);
      break;
  }
  this.sessionWindow.setStatusMessage(statusMsg);
}