const nsIFilePicker = Components.interfaces.nsIFilePicker;
const nsIPrefBranchInternal = Components.interfaces.nsIPrefBranchInternal;

const ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
const prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
const promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
const strBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
const uniConverter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
const winMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);

const PR_UINT32_MAX = 0xffffffff;
const XHTML = "http://www.w3.org/1999/xhtml";

const NS_ERROR_MODULE_NETWORK = 2152398848;
const NS_ERROR_MODULE_FILES = 2152857600;
const NS_ERROR_NET_RESET = NS_ERROR_MODULE_NETWORK + 20;

var Utils = {
  bundle: strBundleService.createBundle("chrome://msnmsgr/locale/common.properties")
};

Utils.askUser = function(aTopic, aQuestion) {
  var flags = promptService.STD_YES_NO_BUTTONS + promptService.BUTTON_POS_1_DEFAULT;
  var check = { value: false };
  var result = promptService.confirmEx(window, aTopic, aQuestion, flags, '', '', '', null, check);
  var answer;
  result == 1 ? answer = false : answer = true;
  return answer;
}

Utils.getExtensionSubDirectory = function(aSubDirectory) {
  var file = Components.classes["@mozilla.org/file/directory_service;1"]
                       .getService(Components.interfaces.nsIProperties)
                       .get("ProfD", Components.interfaces.nsIFile);
  file.append("msnmsgr");
  if (!file.exists() || !file.isDirectory() ) {
    file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0664);
  }
  file.append(aSubDirectory);
  if (!file.exists() || !file.isDirectory() ) {
    file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0664);
  }
  return file;
}

Utils.getMd5HashFromStringAsHex = function(aString) {
  var lengthOut = {}; // out param
  uniConverter.charset = "UTF-8";
  var bytes = uniConverter.convertToByteArray(aString, lengthOut);
  var hash = Components.classes["@mozilla.org/security/hash;1"].createInstance(Components.interfaces.nsICryptoHash);
  hash.init(hash.MD5);
  hash.update(bytes, lengthOut.value);
  var md5bytes = hash.finish(false);
  // the "0"/slice stuff is to ensure all char codes give two hex chars,
  // not just one
  var result = "";
  for (var i = 0; i < md5bytes.length; i++) {
    result += ("0" + md5bytes.charCodeAt(i).toString(16)).slice(-2);
  }
  return result;
}

Utils.getPref = function(aFunction, aPref, aDefaultValue) {
  var prefBranch = prefService.getBranch("extensions.msnmsgr.");
  try {
    return prefBranch[aFunction](aPref);
  }
  catch (aException) {
    return aDefaultValue;
  }
}

Utils.getSha1HashFromFileAsBinary = function(aFile) {
  var fileInputStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
  fileInputStream.init(aFile, 0x01, 0444, 0);
  var hash = Components.classes["@mozilla.org/security/hash;1"].createInstance(Components.interfaces.nsICryptoHash);
  hash.init(hash.SHA1);
  hash.updateFromStream(fileInputStream, PR_UINT32_MAX);
  var sha1bytes = hash.finish(false);
  return sha1bytes;
}

Utils.getSha1HashFromStringAsBase64 = function(aString) {
  var lengthOut = {};
  uniConverter.charset = "UTF-8";
  var bytes = uniConverter.convertToByteArray(aString, lengthOut);
  var hash = Components.classes["@mozilla.org/security/hash;1"].createInstance(Components.interfaces.nsICryptoHash);
  hash.init(hash.SHA1);
  hash.update(bytes, lengthOut.value);
  var sha1string = hash.finish(true);
  return sha1string;
}

Utils.getStatusText = function(aStatusCode) {
  var statusText = this.bundle.GetStringFromName("statusText" + aStatusCode);
  return statusText;
}

Utils.getStringBundle = function(aPath) {
  return strBundleService.createBundle(aPath);
}

Utils.getTimeStamp = function(aSeparator1, aSeparator2) {
  var date = new Date();
  var hours = ensureTwoDigits(date.getHours());
  var minutes = ensureTwoDigits(date.getMinutes());
  var seconds = ensureTwoDigits(date.getSeconds());
  var timeStamp = date.getFullYear()
    + aSeparator1 + (date.getMonth() + 1)
    + aSeparator1 + date.getDate()
    + " " + hours
    + aSeparator2 + minutes
    + aSeparator2 + seconds;

  function ensureTwoDigits(aNumber) {
    var ensuredNumber = "";
    if (aNumber > 9) {
      ensuredNumber = aNumber;
    }
    else {
      ensuredNumber = "0" + aNumber;
    }
    return ensuredNumber;
  }

  return timeStamp;
}

Utils.getUniversallyUniqueId = function() {
  const uuidGenerator = Components.classes["@mozilla.org/uuid-generator;1"].getService(Components.interfaces.nsIUUIDGenerator);
  var uuid = uuidGenerator.generateUUID().toString();
  return uuid;
}

Utils.openSingletonWindow = function(aUrl, aType, aFeatures, aArgument) {
  var recentWindow = winMediator.getMostRecentWindow(aType);
  if (recentWindow) {
    recentWindow.focus();
  }
  else {
    window.openDialog(aUrl, aType, aFeatures, aArgument);
  }
}

Utils.parseHeaders = function(aHeaders) {
  var headers = aHeaders.split("\r\n");
  var parsedHeaders = {};
  for (var header in headers) {
    var parsedHeader = headers[header].split(": ");
    parsedHeaders[parsedHeader[0].toLowerCase()] = parsedHeader[1];
  }
  return parsedHeaders;
}

Utils.convertUTF16BytesToUTF16 = function(aChunk) {
  uniConverter.charset = "UTF-16";
  var utf16String = uniConverter.ConvertToUnicode(aChunk);
  return utf16String;
}

Utils.convertUTF16ToUTF16Bytes = function(aChunk) {
  uniConverter.charset = "UTF-16";
  var utf16Bytes = uniConverter.ConvertFromUnicode(aChunk);
  return utf16Bytes;
}

Utils.convertUTF16ToUTF8Bytes = function(aChunk) {
  uniConverter.charset = "UTF-8";
  var utf8Bytes = uniConverter.ConvertFromUnicode(aChunk);
  return utf8Bytes;
}

Utils.convertUTF8BytesToUTF16 = function(aChunk) {
  uniConverter.charset = "UTF-8";
  var utf16String = uniConverter.ConvertToUnicode(aChunk);
  return utf16String;
}